// C Source File
// Created 09/06/2006; 21:13:23

#include "cri3D.h"
#include <kbd.h>
#include <graph.h>
#include <kbd.h>
#include <stdio.h>
#include <stdlib.h>
#include <compat.h>

/* Internal Variables */
static short transformation_buffer [MAX_POINTS*2] ;
static short sinus_table [] = {0x0000,0x0006,0x000D,0x0013,0x0019,0x001F,0x0025,0x002B,0x0031,0x0037,0x003C,0x0042,0x0047,0x004C,0x0051,0x0056,
	0x005B,0x005F,0x0063,0x0067,0x006A,0x006E,0x0071,0x0074,0x0076,0x0079,0x007A,0x007C,0x007E,0x007F,0x007F,0x0080,
	0x0080,0x0080,0x007F,0x007F,0x007E,0x007C,0x007A,0x0079,0x0076,0x0074,0x0071,0x006E,0x006A,0x0067,0x0063,0x005F,
	0x005B,0x0056,0x0051,0x004C,0x0047,0x0042,0x003C,0x0037,0x0031,0x002B,0x0025,0x001F,0x0019,0x0013,0x000D,0x0006,
	0x0000,0xFFFA,0xFFF3,0xFFED,0xFFE7,0xFFE1,0xFFDB,0xFFD5,0xFFCF,0xFFC9,0xFFC4,0xFFBE,0xFFB9,0xFFB4,0xFFAF,0xFFAA,
	0xFFA5,0xFFA1,0xFF9D,0xFF99,0xFF96,0xFF92,0xFF8F,0xFF8C,0xFF8A,0xFF87,0xFF86,0xFF84,0xFF82,0xFF81,0xFF81,0xFF80,
	0xFF80,0xFF80,0xFF81,0xFF81,0xFF82,0xFF84,0xFF86,0xFF87,0xFF8A,0xFF8C,0xFF8F,0xFF92,0xFF96,0xFF99,0xFF9D,0xFFA1,
	0xFFA5,0xFFAA,0xFFAF,0xFFB4,0xFFB9,0xFFBE,0xFFC4,0xFFC9,0xFFCF,0xFFD5,0xFFDB,0xFFE1,0xFFE7,0xFFED,0xFFF3,0xFFFA,
	0x0000,0x0006,0x000D,0x0013,0x0019,0x001F,0x0025,0x002B,0x0031,0x0037,0x003C,0x0042,0x0047,0x004C,0x0051,0x0056,
	0x005B,0x005F,0x0063,0x0067,0x006A,0x006E,0x0071,0x0074,0x0076,0x0079,0x007A,0x007C,0x007E,0x007F,0x007F,0x0080,
	0x0080,0x0080} ;

#include "cri3DTransform.h"
#include "cri3DFaces.h"
#include "cri3DTexture.h"

void DisplayError (const char * msg) {
  ClrScr () ;
	while (!kbhit()) printf_xy (0, 50, msg);
}

Scene * CreateScene () {
	// Checks the compiler 
	if (sizeof(long) != 4) {
		DisplayError ("The size of a long is different from 4") ;
		exit (-1) ;
	}
	Scene * result = malloc (sizeof(Scene)) ;
	if (result == NULL) {
		DisplayError ("Cannot allocate a new scene") ;
		exit (-1) ;
	}
	ClearScene (result) ;
	return result ;
}

void InitCri3D () {
	// Init internal value
	LCDScreen.address = LCD_MEM ;
	LCDScreen.width = LCD_WIDTH ;
	LCDScreen.height = LCD_HEIGHT ;
	// Internal variable
	unique_access_to_fill_triangle_with_texture = FALSE ;
}

void ReleaseCri3D () {
	//
}

void ClearScene (const Scene * scene) {
	ClearPoints (scene) ;
	ClearConnections (scene) ;
	ClearFaces (scene) ;
}

void AddPoint (const Scene * scene, long x, long y, long z) {
	if (scene->npoints >= MAX_POINTS) {
		DisplayError ("Point limit reached") ;
		exit (-1) ;
	}
	scene->points [scene->npoints * 3 + 0] = x ;
	scene->points [scene->npoints * 3 + 1] = y ;
	scene->points [scene->npoints * 3 + 2] = z ;
	scene->npoints += 1 ;
}

void ClearPoints (const Scene * scene) {
	scene->npoints = 0 ;
}

void AddConnection (const Scene * scene, short point1, short point2) {
	if (scene->nconnections >= MAX_CONNECTIONS) {
		DisplayError ("Connections limit reached") ;
		exit (-1) ;
	}
	scene->connections [scene->nconnections * 2 + 0] = point1 ;
	scene->connections [scene->nconnections * 2 + 1] = point2 ;
	scene->nconnections += 1 ;
}

void ClearConnections (const Scene * scene) {
	scene->nconnections = 0 ;
}

void AddFace (const Scene * scene, short npoints, short points []) {
	AddDecoredFace (scene, -1, npoints, points, NULL) ;
}

void ClearFaces (const Scene * scene) {
	scene->nnfaces = 0 ;
}

void AddDecoredFace (const Scene * scene, const Texture * t, short npoints, short points [], Coordinate mcoords []) {
	int i, index ;
	if (npoints < 3) {
		DisplayError ("A face must have at least 3 points") ;
		exit (-1) ;
	}
	/* Calculates the index */
	index = 0 ;
	for (i = 0 ; i < scene->nnfaces ; i++) {
		index += scene->nfaces [i] ;
	}
	// Check the array can handle it
	if (index + npoints >= MAX_FACES * 3) {
		DisplayError ("Nb decored face limit reached") ;
		exit (-1) ;
	}
	// The amount of points
	scene->nfaces [scene->nnfaces] = npoints ;
	// Copy texture
	scene->textures [scene->nnfaces] = t ;
	// Copy coords and mapping
	for (i = 0 ; i < npoints ; i++) {
		// TODO Check that points [i] < npoints....
		scene->faces [index + i] = points[i] ;
		if (mcoords != NULL)
			scene->mappingcoordinates [index + i] = mcoords [i] ;
	}
	scene->nnfaces += 1;
}

static /* inline */ void RenderConnections (const Screen * screen, const Scene * scene) {
	long i,j=0 ;
	short index1, index2 ;
	WIN_RECT line ;
	PortSet (screen->address, 239, 127) ;
	for (i = 0 ; i < scene->nconnections ; i++) {
		index1 = scene->connections [j++] << 1 ;
		index2 = scene->connections [j++] << 1 ;
		// Draw a line:
		line.x0 = transformation_buffer [index1] ;
		line.y0 = transformation_buffer [index1 + 1] ;
		line.x1 = transformation_buffer [index2] ;
		line.y1 = transformation_buffer [index2 + 1] ;
		DrawClipLine (&line, &(SCR_RECT){{0,0,screen->width, screen->height}}, A_NORMAL) ;
		//CriDrawLine (screen, line.x0, line.y0, line.x1, line.y1) ;
	}
}

static /* inline */ void RenderFaces (const Screen * screen, const Scene * scene) {
	long i, j ;
	short index, npoints ;
	Coordinate coords [] = {{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0}} ;
	Coordinate mcoords [] = {{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0}} ;
	short tmp_index ;
	index = 0 ;
	for (i = 0 ; i < scene->nnfaces ; i++) {
		npoints = scene->nfaces [i] ;
		scene->faces [index] ;
		for (j = 0 ; j < npoints ; j++) {
			tmp_index = scene->faces [index] << 1 ;
			coords [j] = (Coordinate){transformation_buffer [tmp_index], transformation_buffer [tmp_index + 1]} ;
			mcoords [j] = scene->mappingcoordinates [index++] ;
		}
		if (scene->textures [i] == -1) {
			//FillPolygon (screen, npoints,	coords) ;
			if (IsAntiClockwise (scene, i))
				DrawPolygon (screen, npoints, coords) ;
		} else {
			FillPolygonWithTexture (screen, scene->textures [i], npoints,	coords, mcoords) ;
		}
	}
}

void Render (const Screen * screen, const Camera * camera, const Scene * scene) {
	// Render each kind
	RenderConnections (screen, scene) ;
	RenderFaces (screen, scene) ;
}

void Screen2Texture (const Screen * screen, const Texture * texture) {
	void * src, * dst ;
	short width, height ;
	src = screen->address ;
	dst = texture->data ;
	width = texture->width - 1 ;
	if (width <= 0) return ;
	height = texture->height - 1 ;
	if (height <= 0) return ;
#define	MACRO(index) \
	({	\
	asm volatile ("	\
	lsl.b		#1,%%d2 ; \
	bcc			0f ; \
	move.b	#0xff,(%%a1)+ ; \
	jmp			1f ; \
0: ; \
	move.b	#0x00,(%%a1)+ ; \
1: ; \
	"::	\
	) ;	\
	})

	asm volatile ("
	move.w	%[width],%%d0
	move.w	%%d0,%%d4					/* remaining */
	and.w		#7,%%d4
	#sub.w		#1,%%d4
	asr.w		#3,%%d0						/* / 8 for (bit) to (byte) <=> Shift 3 */
	sub.w		#1,%%d0
	move.w	%%d0,%%d3
	move.w	%[height],%%d1
	sub.w		#1,%%d1
	move.l	%[src],%%a0
	move.l	%[dst],%%a1
	move.l	%%a0,%%a2
screen2texture_loop:
	move.b	(%%a0)+,%%d2
	"
	: /* No outputs */
	:	[width] "g" (width), [height] "g" (height), [src] "m" (src), [dst] "m" (dst)) ;
	MACRO (1) ;
	MACRO (2) ;
	MACRO (3) ;
	MACRO (4) ;
	MACRO (5) ;
	MACRO (6) ;
	MACRO (7) ;
	MACRO (8) ;
	asm volatile ("
	dbf			%%d0, screen2texture_loop
	/* Remaing pixels: */
	move.w	%%d4,%%d0
	ble			screen2texture_no_remaining
	move.b	(%%a0)+,%%d2
screen2texture_remaining_pixels:
	lsl.b		#1,%%d2
	bcc			0f
	move.b	#0xff,(%%a1)+
	jmp			1f
0:
	move.b	#0x00,(%%a1)+
1: ;
	dbf			%%d0, screen2texture_remaining_pixels
	/* Next line */
screen2texture_no_remaining:
	move.w	%%d3,%%d0
	lea			30(%%a2),%%a2
	move.l	%%a2,%%a0
	dbf			%%d1, screen2texture_loop
	"
	: /* No outputs */
	: [width] "g" (width)) ;
#undef MACRO
}

short Sinus (short angle) {
	return sinus_table [angle & 0x7f] ;
}

short Cosinus (short angle) {
	return sinus_table [(angle + 32) & 0x7f] ;
}

short GetFaceIndex (const Scene * scene, short face) {
	short i, index ;
	if (face >= scene->nnfaces)
		face = scene->nnfaces ;

	index = 0 ;
	for (i = 0 ; i < face ; i++) {
		index += scene->nfaces [i] ;
	}
	return index ;
}

char IsAntiClockwise (const Scene * scene, short face) {
	short i, index, tmp_index, y_min, index_min, next_index, previous_index, y_max, index_max ;
	Coordinate coords [] = {{0,0},{0,0},{0,0}} ;
	index = GetFaceIndex (scene, face) ;
	//printf_xy (0,50,"n=%d p1=%d p2=%d p3=%d p4=%d", scene->nfaces [face], scene->faces [index], scene->faces [index+1], scene->faces [index+2], scene->faces [index+3]) ;
	// Search y min, and y max
	y_min = 0xfff ;
	y_max = -0xfff ;
	for (i = 0 ; i < scene->nfaces [face] ; i++) {
		tmp_index = scene->faces [index + i] << 1 ;
		coords [i] = (Coordinate){transformation_buffer[tmp_index], transformation_buffer[tmp_index+1]} ;
		if (coords[i].y < y_min) {
			y_min = coords[i].y ;
			index_min = i ;
		}
		if (coords[i].y > y_max) {
			y_max = coords[i].y ;
			index_max = i ;
		}
	}
	next_index = index_min + 1 ;
	if (next_index >= scene->nfaces [face]) next_index = 0 ;
	previous_index = index_min - 1 ;
	if (previous_index < 0) previous_index = scene->nfaces [face] - 1 ;
	return (coords [next_index].x < coords [previous_index].x) ;
	//printf_xy (0,70,"xmed=%d xn=%d xp=%d", coords[index_min].x, coords[next_index].x, coords[previous_index].x) ;
	// If flat triangle, we are unable to not wether it is clockwise or not... so return FALSE !
	/*if (y_min == y_max) return FALSE ;
	if (next_index == index_max) {
		// the third point shoudl be on the right side !
		return (coords[previous_index].x > ((coords[index_min].x + coords[index_max].x) >> 1)) ;
	} else if (previous_index == index_max) {
		// the third point should be on the left side !
		return (coords[next_index].x < ((coords[index_min].x + coords[index_max].x) >> 1)) ;
	} else {
		// Bug in the algorithm
		return FALSE ;
	}*/
}

void SetMappingCoordinate (Scene * scene, short face, Coordinate mcoords []) {
	short i, index ;

	index = GetFaceIndex (scene, face) ;
	for (i = 0 ; i < scene->nfaces [face] ; i++) {
		scene->mappingcoordinates [index + i] = mcoords [i] ;
	}
}

char * Decode (const short width, const short height, const short picture []) {
	char * result, * result_copy ;
	short * picture_copy ;
	short items ;
	char value ;
	long i ;
	short j ;
	unsigned long size ;
	size = width * height ;
	result = malloc (size) ;
	result_copy = result ;
	picture_copy = picture ;
	value = FALSE ;
	for (i = 0 ; i < size ; ) {
		items = *picture_copy ;
		i += items ;
		if (i >= size) items = i - size ;
		for (j = 0 ; j < items ; j++) {
			*result_copy = value ;
			result_copy++ ;
		}
		picture_copy++;
		value = !value ;
	}
	return result ;
}