// C Source File
// Created 09/06/2006; 20:32:56

//#define USE_TI89              // Compile for TI-89
//#define USE_TI92PLUS          // Compile for TI-92 Plus
//#define USE_V200              // Compile for V200

//#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization
#define MIN_AMS 101           // Compile for AMS 1.00 or higher
//#define SAVE_SCREEN           // Save/Restore LCD Contents

#define COMMENT_STRING         "Cube3D by Christophe MOINE"
#define COMMENT_PROGRAM_NAME   "Cube3D"
#define COMMENT_VERSION_STRING "2.0beta"
#define COMMENT_VERSION_NUMBER 2,0,0,0 /* major, minor, revision, subrevision */
#define COMMENT_BW_ICON \
	{0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000}
#define COMMENT_GRAY_ICON \
	{0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000}, \
	{0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000}

//#include <tigcclib.h>
#include <compat.h>
#include <graph.h>
#include <system.h>
#include <kbd.h>
#include <intr.h>
#include <stdio.h>
#include <mem.h>
#include <alloc.h>
//#include "cube.h"
#include "animation.h"
#include "cri3D.h"
#include "textureManager.h"
#include "cubeScene.h"

#define INITIAL_ZOOM_X 180
#define INITIAL_ZOOM_Y 180
#define INITIAL_FOCAL  250

static Texture screenshot_texture ;

short refresh_rate = 0 ;
short frames = 0 ;
#define MAX_MESSAGES 10
char * msgs [MAX_MESSAGES] ;
short durations [MAX_MESSAGES];
short nmsgs ;
#define INSTRUCTION_DURATION 70
#define INSTRUCTION_MESSAGE_1 "Arrows: move cube"
#define INSTRUCTION_MESSAGE_2 "[F1-3]: rendering"
#define INSTRUCTION_MESSAGE_3 "[+-]  : zoom"
#define INSTRUCTION_MESSAGE_4 "[*/]  : perspective"
#define INSTRUCTION_MESSAGE_5 "[F4]  : animation"

static unsigned short arrow_row, symbol_row, row2, row3, row4, row5, esc_row ;

#define ARROWS_ROW 0xfffe      // Constants for low-level keyboard access
#define LEFT_KEY 2
#define RIGHT_KEY 8
#define UP_KEY 1
#define DOWN_KEY 4
#define SECOND_KEY 16
#define SYMBOL_ROW 0xfffd
#define PLUS_KEY 2
#define MINUS_KEY 4
#define STAR_KEY 8
#define SLASH_KEY 16
#define ROW_2   0xfffb
#define T_KEY		32
#define ROW_3   0xfff7
#define ROW_4   0xffef
#define ROW_5   0xffdf
#define F_KEY   128
#define ESC_ROW 0xffbf
#define ESC_KEY 1

#define COUNTER_TIMER_ID 1

CALLBACK void Counter () {
	refresh_rate = frames ;
	frames = 0 ;
}

short GetRefreshRate () {
	return refresh_rate << 1 ;
}

int DisplayMessage (short duration, const char * msg) {
	durations [nmsgs] = duration ;
	msgs [nmsgs++] = msg ;
}

void ClearMessages () {
	nmsgs = 0 ;
}

void PrintMessages (int rowStart) {
	int i ;
	int row ;
	row = rowStart ;
	for (i = 0 ; i < nmsgs ; i++) {
		printf_xy (0, (i + rowStart) * 10, msgs [i]) ;
		if (durations [i] > 0) durations [i] -- ;
	}
	if (i >= 1 && durations [i - 1] <= 0) nmsgs-- ;
}

void TestTexture (const Screen * screen) {
	short tx = 0, ty = 10 ;
	//short picture_width = 50 ;
	// Test texture
	// Draw surrounding:
	FillTriangleWithTexture (screen, &texture,
			&(Coordinate){0+tx,0+ty},&(Coordinate){0,0},
			&(Coordinate){0+tx,40+ty},&(Coordinate){0,20},
			&(Coordinate){40+tx,0+ty},&(Coordinate){20,0}) ;
	FillTriangleWithTexture (screen, &texture,
			&(Coordinate){40+tx,0+ty},&(Coordinate){20,0},
			&(Coordinate){80+tx,40+ty},&(Coordinate){40,20},
			&(Coordinate){80+tx,0+ty},&(Coordinate){40,0}) ;
	FillTriangleWithTexture (screen, &texture,
			&(Coordinate){0+tx,40+ty},&(Coordinate){0,20},
			&(Coordinate){0+tx,80+ty},&(Coordinate){0,40},
			&(Coordinate){40+tx,80+ty},&(Coordinate){20,40}) ;
	FillTriangleWithTexture (screen, &texture,
			&(Coordinate){40+tx,80+ty},&(Coordinate){20,40},
			&(Coordinate){80+tx,80+ty},&(Coordinate){40,40},
			&(Coordinate){80+tx,40+ty},&(Coordinate){40,20}) ;
	// Draw the interior
	FillTriangleWithTexture (screen, &texture,
			&(Coordinate){40+tx,20+ty},&(Coordinate){20,10},
			&(Coordinate){20+tx,40+ty},&(Coordinate){10,20},
			&(Coordinate){40+tx,60+ty},&(Coordinate){20,30}) ;
	FillTriangleWithTexture (screen, &texture,
			&(Coordinate){60+tx,40+ty},&(Coordinate){30,20},
			&(Coordinate){40+tx,20+ty},&(Coordinate){20,10},
			&(Coordinate){40+tx,60+ty},&(Coordinate){20,30}) ;
	
	// Test fill polygon
	tx = 90 ;
	DrawPolygon (screen, 3, &(Coordinate[]) {{tx, 0+ty}, {tx, 40+ty}, {30+tx,0+ty}}) ;
	DrawPolygon (screen, 3, &(Coordinate[]) {{30+tx, 0+ty}, {60+tx,40+ty}, {60+tx,0+ty}}) ;
	DrawPolygon (screen, 3, &(Coordinate[]) {{0+tx, 40+ty}, {tx, 80+ty}, {30+tx,80+ty}}) ;
	DrawPolygon (screen, 3, &(Coordinate[]) {{30+tx, 80+ty}, {60+tx, 80+ty}, {60+tx,40+ty}}) ;
	DrawPolygon (screen, 4, &(Coordinate[]) {{15+tx, 40+ty}, {30+tx, 60+ty}, {45+tx, 40+ty}, {30+tx,20+ty}}) ;
}

void InitIntro () {
	screenshot_texture.width = LCD_HEIGHT ;			// It is not a mistake, this is because width must be < 128
	screenshot_texture.height = LCD_HEIGHT ;
	screenshot_texture.data = malloc (LCD_HEIGHT * LCD_HEIGHT) ;
	if (screenshot_texture.data != NULL) {
		Screen2Texture (&LCDScreen, &screenshot_texture) ;
	}
}

void Intro (const Screen * screen) {
	long tempo ;
	if (screenshot_texture.data) return ;
	Coordinate c11, c12, c13 ;
	Coordinate m11, m12, m13 ;
	Coordinate c21, c22, c23 ;
	Coordinate m21, m22, m23 ;
	Coordinate c31, c32, c33 ;
	Coordinate m31, m32, m33 ;
	c11 = (Coordinate){0,0} ;
	c12 = (Coordinate){0, LCD_HEIGHT} ;
	c13 = (Coordinate){LCD_HEIGHT, LCD_HEIGHT} ;
	c21 = c11 ;
	c22 = (Coordinate){LCD_HEIGHT >> 1, LCD_HEIGHT >> 1} ;
	c23 = (Coordinate){LCD_HEIGHT, 0} ;
	c31 = c22 ;
	c32 = c13 ;
	c33 = c23 ;
	tempo = 0 ;
	PortSet (screen->address, 239, 127) ;
	do {
		// Render
		ClrScr () ;

		// Draw first reactangle
		FillTriangleWithTexture (screen, &screenshot_texture,
			&c11, &(Coordinate){0,0},
			&c12, &(Coordinate){0, LCD_HEIGHT},
			&c13, &(Coordinate){LCD_HEIGHT, LCD_HEIGHT}) ;
		// Draw second rectangle
		FillTriangleWithTexture (screen, &screenshot_texture,
			&c21, &(Coordinate){0,0},
			&c22, &(Coordinate){LCD_HEIGHT >> 1, LCD_HEIGHT >> 1},
			&c23, &(Coordinate){LCD_HEIGHT, 0}) ;
		// Draw last rectangle
		FillTriangleWithTexture (screen, &screenshot_texture,
			&c31, &(Coordinate){LCD_HEIGHT >> 1, LCD_HEIGHT >> 1},
			&c32, &(Coordinate){LCD_HEIGHT, LCD_HEIGHT},
			&c33, &(Coordinate){LCD_HEIGHT, 0}) ;
		
		// Displays the refresh rate
		frames++ ;
		printf_xy (LCD_HEIGHT + 10,0,"%d fps", GetRefreshRate ()) ;

		// Copy the virtual screen to the real screen
		memcpy (LCD_MEM, screen->address, LCD_SIZE);

		// Manages tempo
		if (tempo ++ > 20) break ;
		// Manages the keyboard
		//arrow_row = _rowread (ARROWS_ROW) ;
		//symbol_row = _rowread (SYMBOL_ROW) ;
		//row2 = _rowread (ROW_2) ;
		//row3 = _rowread (ROW_3) ;
		//row4 = _rowread (ROW_4) ;
		//row5 = _rowread (ROW_5) ;
		esc_row = _rowread (ESC_ROW) ;
  } while (!(esc_row & ESC_KEY)) ;
  free (screenshot_texture.data) ;
}


void MainDemo (const Screen * screen)
{
  long     increment ;
  char     decored ;
  char     mode ;
#define CUBE_MODE 0b001
#define TEST_MODE 0b010
#define ANIMATION_MODE 0b100
  Camera   camera ;
  Scene  * scene, * wireframe_scene, * hidden_face_scene, * texture_scene ;

	camera.rx = 0 ;
	camera.ry = 0 ;
	camera.rz = 0 ;
  camera.tx = LCD_WIDTH / 2 ;
	camera.ty = LCD_HEIGHT / 2 ;
	camera.tz = 32 ;
	camera.zoomx = INITIAL_ZOOM_X ;
	camera.zoomy = INITIAL_ZOOM_Y ;
	camera.focal = INITIAL_FOCAL ;

	// Build Wireframe scene
  wireframe_scene = CreateScene () ;
	FillPoints (wireframe_scene) ;
	FillConnections (wireframe_scene) ;
	// Build Hidden faces scene
	hidden_face_scene = CreateScene () ;
	FillPoints (hidden_face_scene) ;
	FillFaces (hidden_face_scene) ;
	// Build Texture scene
	texture_scene = CreateScene () ;
	FillPoints (texture_scene) ;
	FillDecoredFaces (texture_scene) ;
	// Initial render is wireframe mode
	increment = 1 ;
	scene = wireframe_scene ;
	decored = FALSE ;
	mode = CUBE_MODE ;
	InitAnimation () ;
	PortSet (screen->address, 239, 127) ;
	do {
		// Render
		ClrScr () ;

		// Render
		if (mode == TEST_MODE) {
			TestTexture (screen) ;
		} else if (mode == ANIMATION_MODE) {
			Animate (screen) ;
		} else if (mode == CUBE_MODE) {
			// Update textures if needed
			if (decored) UpdateTextures (scene, &camera) ;
			// transformation
			Transform (&camera, scene) ;
			// Final rendering
			Render (screen, &camera, scene) ;
		}
		// Displays the refresh rate
		PortSet (screen->address, 239, 127) ;
		frames++ ;
		printf_xy (0,0,"%d fps", GetRefreshRate ()) ;
		PrintMessages (1) ;
		// Copy the virtual screen to the real screen
		memcpy (LCD_MEM, screen->address, LCD_SIZE);
		// Manages the keyboard
		arrow_row = _rowread (ARROWS_ROW) ;
		symbol_row = _rowread (SYMBOL_ROW) ;
		row2 = _rowread (ROW_2) ;
		row3 = _rowread (ROW_3) ;
		row4 = _rowread (ROW_4) ;
		row5 = _rowread (ROW_5) ;
		esc_row = _rowread (ESC_ROW) ;
		if (mode == CUBE_MODE) {
			if (arrow_row & UP_KEY) {
				camera.rx += increment ;
			}
			if (arrow_row & DOWN_KEY) {
				camera.rx -= increment ;
			}
			if (arrow_row & LEFT_KEY) {
				camera.ry += increment ;
			}
			if (arrow_row & RIGHT_KEY) {
				camera.ry -= increment ;
			}
			if (symbol_row & PLUS_KEY) {
				camera.zoomx += 5 ;
				camera.zoomy += 5 ;
			}
			if (symbol_row & MINUS_KEY) {
				camera.zoomx -= 5 ;
				camera.zoomy -= 5 ;
			}
			if (symbol_row & STAR_KEY) {
				camera.focal += 10 ;
			}
			if (symbol_row & SLASH_KEY) {
				camera.focal -= 10 ;
			}
		}
		if (row5 & F_KEY) { /* F1 */
			scene = wireframe_scene ;
			mode = CUBE_MODE ;
  		increment = 1 ;
  		decored = FALSE ;
		}
		if (row4 & F_KEY) { /* F2 */
			scene = hidden_face_scene ;
			mode = CUBE_MODE ;
  		increment = 1 ;
  		decored = FALSE ;
		}
		if (row3 & F_KEY) { /* F3 */
			scene = texture_scene ;
			mode = CUBE_MODE ;
  		increment = 4 ;
  		decored = TRUE ;
		}
		if (row2 & F_KEY) { /* F4 */
			mode = ANIMATION_MODE ;
			ClearMessages () ;
		}
		if (row2 & T_KEY) { /* t */
			mode = TEST_MODE ;
			ClearMessages () ;
		}
  } while (!(esc_row & ESC_KEY)) ;
  ReleaseAnimation () ;
	free (wireframe_scene) ;
  free (hidden_face_scene) ;
  free (texture_scene) ;
}

// Main Function
void _main(void)
{
  Screen   screen ;
	screen.width = LCD_WIDTH ;
	screen.height = LCD_HEIGHT ;
	screen.address = malloc (LCD_SIZE);  // Allocate the buffer
	if (!screen.address) {
		printf_xy (0,0,"Cannot allocate mem for virtual screen") ;
		printf_xy (0,10,"Program will terminate") ;
	} else {
	  InitCri3D () ;
		InitTextures () ;
		InitIntro () ;

		OSSetSR(0x0400);
		OSVRegisterTimer (COUNTER_TIMER_ID, 10, Counter) ;

		// Intro has not been finished yet
		//Intro (&screen) ;

		DisplayMessage (INSTRUCTION_DURATION, INSTRUCTION_MESSAGE_1) ;
		DisplayMessage (INSTRUCTION_DURATION, INSTRUCTION_MESSAGE_2) ;
		DisplayMessage (INSTRUCTION_DURATION, INSTRUCTION_MESSAGE_3) ;
		DisplayMessage (INSTRUCTION_DURATION, INSTRUCTION_MESSAGE_4) ;
		DisplayMessage (INSTRUCTION_DURATION, INSTRUCTION_MESSAGE_5) ;
	  MainDemo (&screen) ;
	
	  // Release stuffs
	  ReleaseCri3D () ;
	  ReleaseTextures () ;
	  PortRestore () ;
	  //CriPortRestore () ;
	  OSVFreeTimer (COUNTER_TIMER_ID) ;
	  OSSetSR(0x0000) ;
	  free (screenshot_texture.data) ;
	  free (screen.address) ;
  }
}
