// C Source File
// Created 28/06/2006; 16:57:58

#include "textureManager.h"
#include "at_picture.h"
#include "spiral_picture.h"
#include "cubeScene.h"
#include "duck_picture_1.h"
#include "duck_picture_2.h"
#include "duck_picture_3.h"
#include "duck_picture_4.h"
#include "duck_picture_5.h"

#include <graph.h>
#include <alloc.h>
#include <error.h>
#include <stdlib.h>
#include <string.h>
#include <compat.h>
#include <math.h>
// Just for tests
#include <stdio.h>
#include <mem.h>

static Camera cube_camera ;
static Scene * cube_scene ;
static Screen cube_screen ;
static short frame ;
static Screen about_screen ;
static Screen ball_screen ;

#define NB_MESSAGES 8
static char * about_message [] = {"Cube", "Demo", "C. MOINE", "Cri3D", "engine", "50%% C", "50%% 68k", ""} ;

#define ABOUT_TEXTURE 0
#define ROTATE_TEXTURE 4
#define ZOOM_TEXTURE 5
#define ANIMATED_TEXTURE 3
#define CUBE_TEXTURE 2
#define BALL_TEXTURE 1

#define INSTRUCTION_DURATION 70
#define INSTRUCTION_MESSAGE_BALL_1 "Use Up and Down keys"
#define INSTRUCTION_MESSAGE_BALL_2 "to move the ball"

void InitTextures () {
	short i, j, k ;
	texture.width = AT_WIDTH ;
	texture.height = AT_HEIGHT ;
	texture.data = Decode (AT_WIDTH, AT_HEIGHT, at_header_data) ;
	// The screenshot is used as texture
	textures[BALL_TEXTURE].width = BALL_WIDTH ;
	textures[BALL_TEXTURE].height = BALL_HEIGHT ;
	ball_screen.width = BALL_WIDTH ;
	ball_screen.height = BALL_HEIGHT ;
	ball_screen.address = malloc (LCD_SIZE) ;
	if (ball_screen.address != NULL) {
		textures[BALL_TEXTURE].data = malloc (BALL_WIDTH * BALL_HEIGHT) ;
	} else {
		textures[BALL_TEXTURE].data = NULL ;
	}
	// About texture
	textures[ABOUT_TEXTURE].width = 100 ;
	textures[ABOUT_TEXTURE].height = 200 ;
	about_screen.width = ABOUT_WIDTH ;
	about_screen.height = ABOUT_HEIGHT ;
	about_screen.address = malloc (LCD_SIZE) ;
	if (about_screen.address) {
		PortSet (about_screen.address, 239, 200) ;
		for (i = 0 ; i < NB_MESSAGES ; i++) {
			printf_xy ((ABOUT_WIDTH >> 1)- ((strlen (about_message[i]) >> 1) * 5),i*10,about_message[i]) ;
		}
		for (j = 0 ; j < (ABOUT_HEIGHT / 10) ; j++) {
			printf_xy ((ABOUT_WIDTH >> 1)- ((strlen (about_message[j]) >> 1) * 5),(i+j)*10,about_message[j]) ;
		}
		textures[ABOUT_TEXTURE].data = malloc (100 * 200);
		Screen2Texture (&about_screen, &textures[0]) ;
	} else {
		textures[ABOUT_TEXTURE].data = NULL ;
	}
	// Wireframe cube
	cube_camera.rx = 0 ;
	cube_camera.ry = 0 ;
	cube_camera.rz = 0 ;
	cube_camera.tx = CUBE_WIDTH / 2 ;
	cube_camera.ty = CUBE_HEIGHT / 2 ;
	cube_camera.tz = 50 ;
	cube_camera.zoomx = 100 ;
	cube_camera.zoomy = 100 ;
	cube_camera.focal = 200 ;
	cube_scene = CreateScene () ;
	FillPoints (cube_scene) ;
	FillConnections (cube_scene) ;
	cube_screen.width = CUBE_WIDTH ;
	cube_screen.height = CUBE_HEIGHT ;
	cube_screen.address = malloc (LCD_SIZE) ;
	if (cube_screen.address != NULL) {
		textures[CUBE_TEXTURE].width = CUBE_WIDTH ;
		textures[CUBE_TEXTURE].height = CUBE_HEIGHT ;
		textures[CUBE_TEXTURE].data = malloc (CUBE_WIDTH * CUBE_HEIGHT);
	} else {
		textures[CUBE_TEXTURE].data = NULL ;
	}
	// Animated texture (duck)
	frame = 0 ;
	textures[ANIMATED_TEXTURE].width = DUCK_WIDTH ;
	textures[ANIMATED_TEXTURE].height = DUCK_HEIGHT ;
	textures[ANIMATED_TEXTURE].data = duck1_header_data ;
	// Zoom In/Out
	textures[ZOOM_TEXTURE].width = AT_WIDTH ;
	textures[ZOOM_TEXTURE].height = AT_HEIGHT ;
	textures[ZOOM_TEXTURE].data = Decode (AT_WIDTH, AT_HEIGHT, at_header_data) ;
	// Rotate
	textures[ROTATE_TEXTURE].width = SPIRAL_WIDTH ;
	textures[ROTATE_TEXTURE].height = SPIRAL_HEIGHT ;
	textures[ROTATE_TEXTURE].data = Decode (SPIRAL_WIDTH, SPIRAL_HEIGHT, spiral_header_data) ;
}

void UpdateCubeTexture (short face) {
	int x, y, index_src, index_dst, byte_count ;
	long tempo ;
	char src ;
	// Texture for the cube
	cube_camera.rx += CUBE_ROTATION_INCREMENT ;
	cube_camera.ry += CUBE_ROTATION_INCREMENT ;
	cube_camera.rz += CUBE_ROTATION_INCREMENT ;
	// Render
	PortSet (cube_screen.address, 239, 127) ;
	ClrScr () ;
	Transform (&cube_camera, cube_scene) ;
	Render (&cube_screen, &cube_camera, cube_scene) ;
	// Transfer the screen buffer, to the texture (cube header data)
	Screen2Texture (&cube_screen,&textures[face]) ;
}

void UpdateDuckTexture (short face) {
	switch (frame >> 1) {
		case 0: textures[face].data = duck1_header_data ;
						break ;
		case 1: textures[face].data = duck2_header_data ;
						break ;
		case 2: textures[face].data = duck3_header_data ;
						break ;
		case 3: textures[face].data = duck4_header_data ;
						break ;
		case 4: textures[face].data = duck5_header_data ;
						break ;
	}
	frame++ ;
	if ((frame >> 1) >= 5) {
		frame = 0 ;
	}
}

void UpdateZoomTexture (Scene * scene, short face) {
	static short a = 0 ;
	short  delta ;
	//Coordinate * c ;
	delta = (Sinus (a) >> ZOOM_DELTA_BIT_SHIFT) + ZOOM_DELTA ;
	a += ZOOM_INCREMENT ;
	SetMappingCoordinate (scene, face, (Coordinate []){{0 + delta, 0 + delta},
			{0 + delta, ZOOM_HEIGHT - delta},
			{ZOOM_WIDTH - delta, ZOOM_HEIGHT - delta},
			{ZOOM_WIDTH - delta, 0 + delta}}) ;
}

short inline GetRotationX (short angle) {
	short cos ;
	cos = Cosinus (angle) ;
	return ((cos << ROTATION_CIRCLE_RADIUS_FACTOR_1) + (cos << ROTATION_CIRCLE_RADIUS_FACTOR_2) >> 7) ;
}

short inline GetRotationY (short angle) {
	short sin ;
	sin = Sinus (angle) ;
	return ((sin << ROTATION_CIRCLE_RADIUS_FACTOR_1) + (sin << ROTATION_CIRCLE_RADIUS_FACTOR_2) >> 7) ;
}

void UpdateRotateTexture (Scene * scene, short face) {
	static short a = 0 ;
	Coordinate * c ;

	a += ROTATION_INCREMENT ;
	/* Exact formula:
	SetMappingCoordinate (scene, 5, (Coordinate []){
			{((ROTATION_CIRCLE_RADIUS * Cosinus (a)) >> 7) + ROTATION_CENTER_X, ((ROTATION_CIRCLE_RADIUS * Sinus (a)) >> 7) + ROTATION_CENTER_Y},
			{((ROTATION_CIRCLE_RADIUS * Cosinus (a + 32)) >> 7) + ROTATION_CENTER_X, ((ROTATION_CIRCLE_RADIUS * Sinus (a + 32)) >> 7) + ROTATION_CENTER_Y},
			{((ROTATION_CIRCLE_RADIUS * Cosinus (a + 64)) >> 7) + ROTATION_CENTER_X, ((ROTATION_CIRCLE_RADIUS * Sinus (a + 64)) >> 7) + ROTATION_CENTER_Y},
			{((ROTATION_CIRCLE_RADIUS * Cosinus (a + 94)) >> 7) + ROTATION_CENTER_X, ((ROTATION_CIRCLE_RADIUS * Sinus (a + 94)) >> 7) + ROTATION_CENTER_Y}}) ;
		Replaced into an approximative formula (much faster, by using bit shifting)
			*/
	SetMappingCoordinate (scene, face, (Coordinate []){
			{GetRotationX (a) + ROTATION_CENTER_X, GetRotationY (a) + ROTATION_CENTER_Y},
			{GetRotationX (a + 32) + ROTATION_CENTER_X, GetRotationY (a + 32) + ROTATION_CENTER_Y},
			{GetRotationX (a + 64) + ROTATION_CENTER_X, GetRotationY (a + 64) + ROTATION_CENTER_Y},
			{GetRotationX (a + 96) + ROTATION_CENTER_X, GetRotationY (a + 96) + ROTATION_CENTER_Y}}) ;
}

void UpdateAboutTexture (const Scene * scene, const Camera * camera, int face) {
	static short ty = 0 ;
	long rotation ;
	rotation = camera->ry & 0xff ;
	ty ++ ;
	if (ty > NB_MESSAGES * 10) ty = 0 ;
	
	// Avoid the text to be upside down
	if (Cosinus (camera->ABOUT_MOVE_AXIS) < 0) {
		SetMappingCoordinate (scene, face, (Coordinate []){
				{ABOUT_WIDTH,0 + ty},
				{0,0 + ty},
				{0,ABOUT_HEIGHT + ty},
				{ABOUT_WIDTH,ABOUT_HEIGHT + ty}}) ;
	} else {
		SetMappingCoordinate (scene, face, (Coordinate []){
				{0,ABOUT_HEIGHT + ty},
				{ABOUT_WIDTH,ABOUT_HEIGHT + ty},
				{ABOUT_WIDTH,0 + ty},
				{0,0 + ty}}) ;
	}
}

void UpdateBallTexture (const Scene * scene, const Camera * camera, int face) {
	static short vx = 0, vy = 0 ;
	static short x = BALL_WIDTH / 2, y = BALL_HEIGHT / 2 ;
	static long old_rotation = 0xffff ;
	static char first_rendering = TRUE ;
	if (first_rendering) {
		DisplayMessage (INSTRUCTION_DURATION, INSTRUCTION_MESSAGE_BALL_1) ;
		DisplayMessage (INSTRUCTION_DURATION, INSTRUCTION_MESSAGE_BALL_2) ;
		first_rendering = FALSE ;
	}
	/*static short message_timer = 150 ;
	static long old_rx = 0xffff ;
	// Instruction message
	if (message_timer-- > 0 && (old_rx == 0xffff || camera->rx == old_rx)) {
		printf_xy (0,10, "Move the ball") ;
		printf_xy (0,20, "Using [up]/[down] keys !") ;
	} else {
		message_timer = 0 ;
	}
	old_rx = camera->rx ;*/
	// Calculate the ball position
	x += vx >> 7 ;
	y += vy >> 7 ;
	// Considers the cube movements
	/*if (old_rotation != 0xffff && old_rotation != camera->BALL_MOVE_AXIS) {
		float r ;
		float a ;
		float xf ;
		float yf ;
		float delta ; // Delta for the angle
		//delta = ((float)(camera->BALL_MOVE_AXIS - old_rotation)) * 360.0 / 128.0 ;
		xf = (((float)x) / 128.0) - (float)(BALL_WIDTH / 2) ;
		yf = (((float)y) / 128.0) - (float)(BALL_HEIGHT / 2) ;
		r = sqrt ((xf * xf) + (yf * yf)) ;
		a = 2.0 * atan ((yf / r) / (1.0 + (xf / r))) ;
		x = ((BALL_WIDTH / 2) << 7) + (short)(r * cos (a) * 128.0) ;
		y = ((BALL_HEIGHT / 2) << 7) + (short)(r * sin (a) * 128.0) ;
		//x = (BALL_WIDTH / 2) + (((short)r * Cosinus (camera->BALL_MOVE_AXIS + (short)a)) >> 7) ;
		//y = (BALL_HEIGHT / 2) + (((short)r * Sinus (camera->BALL_MOVE_AXIS + (short)a)) >> 7) ;
	}
	old_rotation = camera->BALL_MOVE_AXIS ;*/
	// Manages borders
	if (x < BALL_RADIUS) {
		vx = -vx ;
		x = BALL_RADIUS ;
	}
	if (x >= BALL_WIDTH - BALL_RADIUS) {
		vx = -vx ;
		x = BALL_WIDTH - BALL_RADIUS ;
	}
	if (y < BALL_RADIUS) {
		vy = -vy ;
		y = BALL_RADIUS ;
	}
	if (y >= BALL_HEIGHT - BALL_RADIUS) {
		vy = -vy ;
		y = BALL_HEIGHT - BALL_RADIUS ;
	}
	// Considers the gravity
	vx -= Sinus (camera->BALL_MOVE_AXIS) ;
	vy += Cosinus (camera->BALL_MOVE_AXIS) ;
	// Considers the frictions
	vx -= vx >> BALL_FRICTION_BIT_SHIFT ;
	//if (vx < BALL_SPEED_TRIGGER) vx = 0 ;
	vy -= vy >> BALL_FRICTION_BIT_SHIFT ;
	//if (vy < BALL_SPEED_TRIGGER) vy = 0 ;
	// Draw the ball
	PortSet (ball_screen.address, 239, 127) ;
	ClrScr () ;

	DrawLine (0,0,BALL_WIDTH-1,0,A_NORMAL) ;
	DrawLine (BALL_WIDTH-1,0,BALL_WIDTH-1,BALL_HEIGHT-2,A_NORMAL) ;
	DrawLine (BALL_WIDTH-1,BALL_HEIGHT-2,0,BALL_HEIGHT-2,A_NORMAL) ;
	DrawLine (0,BALL_HEIGHT-1,0,0,A_NORMAL) ;
	DrawClipEllipse (x, y,
		BALL_RADIUS, BALL_RADIUS,
		&(SCR_RECT){{0, 0, BALL_WIDTH, BALL_HEIGHT}},
		A_NORMAL) ;
	Screen2Texture (&ball_screen, &textures[face]) ;
}

void UpdateTextures (const Scene * scene, const Camera * camera) {
	//if (IsAntiClockwise (scene, 3)) {
	UpdateDuckTexture (ANIMATED_TEXTURE) ;
	//}
	//if (IsAntiClockwise (scene, 5)) {
	UpdateZoomTexture (scene, ZOOM_TEXTURE) ;
	//}
	if (IsAntiClockwise (scene, ROTATE_TEXTURE)) {
		UpdateRotateTexture (scene, ROTATE_TEXTURE) ;
	}
	// Cube is the last, because it executes PortSet (side effects)
	if (IsAntiClockwise (scene, CUBE_TEXTURE)) {
		UpdateCubeTexture (CUBE_TEXTURE) ;
	}
	//if (IsAntiClockwise (scene, ABOUT_TEXTURE)) {
	UpdateAboutTexture (scene, camera, ABOUT_TEXTURE) ;
	//}
	if (IsAntiClockwise (scene, BALL_TEXTURE)) {
		UpdateBallTexture (scene, camera, BALL_TEXTURE) ;
	}
}

void ReleaseTextures () {
	if (texture.data != NULL) {
		free (texture.data) ;
	}
	if (textures[ABOUT_TEXTURE].data != NULL) {
		free (textures[ABOUT_TEXTURE].data) ;
	}
	if (about_screen.address != NULL) {
		free (about_screen.address) ;
	}
	if (textures[BALL_TEXTURE].data != NULL) {
		free (textures[BALL_TEXTURE].data) ;
	}
	if (ball_screen.address != NULL) {
		free (ball_screen.address) ;
	}
	if (cube_screen.address != NULL) {
		free (cube_screen.address) ;
	}
	if (textures[CUBE_TEXTURE].data != NULL) {
		free (textures[CUBE_TEXTURE].data) ;
	}
	if (cube_scene != NULL) {
		free (cube_scene) ;
	}
	if (textures[ROTATE_TEXTURE].data != NULL) {
		free (textures[ROTATE_TEXTURE].data) ;
	}
	if (textures[ZOOM_TEXTURE].data != NULL) {
		free (textures[ZOOM_TEXTURE].data) ;
	}
}